<?php
namespace Zeedhi\Tools\DatabaseReader\Adapters;

use Doctrine\DBAL\Connection;
use Zeedhi\Tools\DatabaseReader\AdapterReader;
use Zeedhi\Tools\DatabaseReader\DTO;
use Zeedhi\Tools\DatabaseReader\Exception;

class PostgreSQL extends AdapterReader{

    protected static $deleteRuleConversion = array(
        'CASCADE'   => DTO\Constraints\ForeignKey::DELETE_RULE_CASCADE,
        'NO ACTION' => DTO\Constraints\ForeignKey::DELETE_RULE_NO_ACTION,
        'RESTRICT'  => DTO\Constraints\ForeignKey::DELETE_RULE_RESTRICT,
        'SET NULL'  => DTO\Constraints\ForeignKey::DELETE_RULE_SET_NULL
    );

    const CONSTRAINT_TYPE_PK = "p";
    const CONSTRAINT_TYPE_FK = "f";
    const CONSTRAINT_TYPE_UNIQUE = "u";

    protected static $columnDataTypeConversion = array(
        'bytea'               => DTO\Column::DATA_TYPE_BINARY,
        'blob'                => DTO\Column::DATA_TYPE_BINARY,
        'binary large'        => DTO\Column::DATA_TYPE_BINARY,

        'boolean'             => DTO\Column::DATA_TYPE_BOOLEAN,
        'bool'                => DTO\Column::DATA_TYPE_BOOLEAN,
        'bit'                 => DTO\Column::DATA_TYPE_BOOLEAN,

        'smallint'            => DTO\Column::DATA_TYPE_INTEGER,
        'int2'                => DTO\Column::DATA_TYPE_INTEGER,
        'integer'             => DTO\Column::DATA_TYPE_INTEGER,
        'int'                 => DTO\Column::DATA_TYPE_INTEGER,
        'int4'                => DTO\Column::DATA_TYPE_INTEGER,
        'bigint'              => DTO\Column::DATA_TYPE_INTEGER,
        'int8'                => DTO\Column::DATA_TYPE_INTEGER,
        'smallserial'         => DTO\Column::DATA_TYPE_INTEGER,
        'serial2'             => DTO\Column::DATA_TYPE_INTEGER,
        'serial'              => DTO\Column::DATA_TYPE_INTEGER,
        'serial4'             => DTO\Column::DATA_TYPE_INTEGER,
        'bigserial'           => DTO\Column::DATA_TYPE_INTEGER,
        'serial8'             => DTO\Column::DATA_TYPE_INTEGER,

        'decimal'             => DTO\Column::DATA_TYPE_FLOAT,
        'numeric'             => DTO\Column::DATA_TYPE_FLOAT,
        'real'                => DTO\Column::DATA_TYPE_FLOAT,
        'float4'              => DTO\Column::DATA_TYPE_FLOAT,
        'double precision'    => DTO\Column::DATA_TYPE_FLOAT,
        'float8'              => DTO\Column::DATA_TYPE_FLOAT,
        'money'               => DTO\Column::DATA_TYPE_FLOAT,

        'char'                => DTO\Column::DATA_TYPE_STRING,
        'character'           => DTO\Column::DATA_TYPE_STRING,

        'varbit'              => DTO\Column::DATA_TYPE_STRING,
        'bit varying'         => DTO\Column::DATA_TYPE_STRING,
        'varchar'             => DTO\Column::DATA_TYPE_STRING,
        'character varying'   => DTO\Column::DATA_TYPE_STRING,
        'text'                => DTO\Column::DATA_TYPE_STRING,

        'date'                => DTO\Column::DATA_TYPE_DATE,
        'time'                => DTO\Column::DATA_TYPE_DATETIME,
        'timetz'              => DTO\Column::DATA_TYPE_DATETIME,
        'timestamptz'         => DTO\Column::DATA_TYPE_DATETIME
    );

    /** @var \Doctrine\DBAL\Connection */
    protected $connection;

    public function __construct(Connection $connection) {
        $this->connection = $connection;
    }

    /**
     * Get name from all tables in database.
     *
     * @return string[] With name of all tables.
     */
    public function getAllTableNames() {
        $tableRows = $this->connection->fetchAll("select table_name from information_schema.tables where table_schema = 'public'");
        $tableNames = array();
        foreach($tableRows as $dbTable){
            $tableNames[] = $dbTable['table_name'];
        }

        return $tableNames;
    }

    /**
     * Get a table DTObject with all data about chosen table.
     *
     * @param string $tableName Name from chosen table.
     *
     * @return DTO\Table
     *
     * @throws \Zeedhi\Tools\DatabaseReader\Exception When table was not found in database.
     */
    public function getTable($tableName) {
        $tableRow = $this->getTableRow($tableName);
        $DTOTable = new DTO\Table();
        $DTOTable->setName($tableRow['table_name']);
        return $DTOTable;
    }

    /**
     * Return constraints rows of given table that belongs to chosen table.
     *
     * @param string $tableName      Name of table that belongs constraints.
     * @param string $constraintType Type of constraint, must be on of self::CONSTRAINT_TYPE_*.
     *
     * @return array of Rows that are array too.
     */
    protected function getConstraintRow($tableName, $constraintType) {
        static $statement = null;
        if ($statement === null) {
            $statement = $this->connection->prepare("select pg_constraint.conname, pg_constraint.confdeltype from pg_constraint join pg_class on pg_constraint.conrelid = pg_class.oid
                                                     where pg_constraint.contype=:CONSTRAINT_TYPE and pg_class.relname=:TABLE_NAME");
        }

        $statement->execute(array(":TABLE_NAME" => $tableName, "CONSTRAINT_TYPE" => $constraintType));
        return $statement->fetchAll();
    }

    protected function getPrimaryKeyRow($tableName) {
        $primaryKeyRow = $this->getConstraintRow($tableName, self::CONSTRAINT_TYPE_PK);
        if (!$primaryKeyRow) throw Exception::tableWithoutPrimaryKey($tableName);
        return $primaryKeyRow;
    }

    protected function getForeignKeyRows($tableName) {
        static $statement = null;
        if ($statement === null) {
            $statement = $this->connection->prepare("SELECT DISTINCT c.CONSTRAINT_NAME, c.UNIQUE_CONSTRAINT_NAME, c.DELETE_RULE
                                                                     FROM information_schema.key_column_usage k
                                                                     INNER JOIN information_schema.referential_constraints c
                                                                     ON c.constraint_name = k.constraint_name
                                                                     WHERE k.table_name = :TABLE_NAME");
        }
        $statement->execute(array(":TABLE_NAME" => $tableName));
        return $statement->fetchAll();
    }

    protected function getUniqueConstraintRows($tableName) {
        return $this->getConstraintRow($tableName, self::CONSTRAINT_TYPE_UNIQUE);
    }

    protected function getDTOForeignKeyFromRow($foreignKeyRow){
        if(!isset(self::$deleteRuleConversion[$foreignKeyRow['DELETE_RULE']])) {
            throw Exception::unsupportedDeleteRule($foreignKeyRow['DELETE_RULE']);
        }
        $deleteRule = self::$deleteRuleConversion[$foreignKeyRow['DELETE_RULE']];
        $DTOForeignKey = new DTO\Constraints\ForeignKey(
            $foreignKeyRow['CONSTRAINT_NAME'],
            $foreignKeyRow['UNIQUE_CONSTRAINT_NAME'],
            $deleteRule
        );
        return $DTOForeignKey;
    }

    protected function getDTOUniqueConstraintsFromRows($uniqueConstraintRows) {
        $DTOUniqueConstraint = array();
        foreach($uniqueConstraintRows as $uniqueConstraintRow) {
            $DTOUniqueConstraint[] = new DTO\Constraints\Unique($uniqueConstraintRow['conname']);
        }

        return $DTOUniqueConstraint;
    }

    /**
     * @param $tableName
     *
     * @return mixed
     *
     * @throws \Zeedhi\Tools\DatabaseReader\Exception
     */
    protected function getTableRow($tableName) {
        static $statement = null;
        if ($statement === null) {
            $statement = $this->connection->prepare("select table_name from information_schema.tables where table_schema = 'public' and table_name=:TABLE_NAME");
        }
        $statement->execute(array('TABLE_NAME' => $tableName));
        $tableRow = $statement->fetch(\PDO::FETCH_ASSOC);
        if (!$tableRow) throw Exception::tableNotFound($tableName);
        return $tableRow;
    }

    /**
     * @param $tableName
     * @return array
     */
    protected function getColumnRows($tableName) {
        static $statement = null;
        if ($statement === null) {
            $statement = $this->connection->prepare("SELECT column_name,data_type,character_maximum_length, numeric_precision, numeric_scale, is_nullable, column_default FROM information_schema.columns WHERE table_name = :TABLE_NAME");
        }

        $statement->execute(array('TABLE_NAME' => $tableName));
        $columnsRow = $statement->fetchAll();
        return $columnsRow;
    }

    /**
     * Factory a DTO\Column from a row of postgreSQL Dictionary.
     *
     * @param array $columnRow With positions COLUMN_NAME, DATA_DEFAULT, DATA_LENGTH, DATA_PRECISION, DATA_SCALE, DATA_TYPE, NULLABLE
     *
     * @return DTO\Column
     */
    protected function getDTOColumnFromRow($columnRow) {
        $DTOColumn = new DTO\Column();
        $DTOColumn->setName($columnRow['column_name']);
        $DTOColumn->setDataType(self::$columnDataTypeConversion[$columnRow['data_type']]);
        $DTOColumn->setDataDefault($columnRow['column_default']);
        $DTOColumn->setDataLength($columnRow['character_maximum_length']);
        $DTOColumn->setDataPrecision($columnRow['numeric_precision']);
        $DTOColumn->setDataScale($columnRow['numeric_scale']);
        $DTOColumn->setNullAble($columnRow['is_nullable'][0]);
        return $DTOColumn;
    }

    /**
     * @param $primaryKeyRow
     * @return DTO\Constraints\PrimaryKey
     */
    protected function getDTOPrimaryKeyFromRow($primaryKeyRow) {
        $DTOPrimaryKey = new DTO\Constraints\PrimaryKey($primaryKeyRow['conname']);
        return $DTOPrimaryKey;
    }

    protected function getDTOConsColumnFromRow($consColumnRow) {
        $DTOConsColumn = new DTO\Constraints\ConsColumn($consColumnRow['column_name'], $consColumnRow['ordinal_position']);
        return $DTOConsColumn;
    }

    protected function getConsColumnsRow($constraintName) {
        static $statement = null;
        if ($statement === null) {
            $statement = $this->connection->prepare("select column_name, ordinal_position from information_schema.key_column_usage where constraint_name=:CONSTRAINT_NAME");
        }

        $statement->execute(array(":CONSTRAINT_NAME" => $constraintName));
        return $statement->fetchAll();
    }

    /**
     * Get foreign key from chosen table with given name.
     *
     * @param string $tableName Name from chosen table.
     * @param string $constraintName Name of constraint.
     *
     * @throws \Zeedhi\Tools\DatabaseReader\Exception When table was not found in database.
     *
     * @return DTO\Constraints\ForeignKey
     */
    public function getForeignKeyByNameAndTableName($constraintName, $tableName){
        $statement = $this->connection->prepare("SELECT DISTINCT c.CONSTRAINT_NAME, c.UNIQUE_CONSTRAINT_NAME, c.DELETE_RULE
                                                                     FROM information_schema.key_column_usage k
                                                                     INNER JOIN information_schema.referential_constraints c
                                                                     ON c.constraint_name = k.constraint_name
                                                                     WHERE k.table_name = :TABLE_NAME AND c.CONSTRAINT_NAME = :CONSTRAINT_NAME");
        $statement->execute(array(':TABLE_NAME' => $tableName, ':CONSTRAINT_NAME' => $constraintName));
        $constraintRow = $statement->fetch(\PDO::FETCH_ASSOC);
        if (!$constraintRow) throw Exception::constraintNotFoundInDatabase($tableName, $constraintName);

        $DTOForeignKey = $this->getDTOForeignKeyFromRow($constraintRow);
        $this->completeConsColumnForConstraint($DTOForeignKey);
        return $DTOForeignKey;
    }

    /**
     * Get the sequence of that table.that sequence.
     *
     *
     * @param string $tableName Name of table.
     * @param string $columnName Name of column that uses
     * @return DTO\Sequence
     */
    public function getSequenceFromTableAndColumn($tableName, $columnName){
        $seqNameStatement = $this->connection->prepare("SELECT sequence_name, start_value, minimum_value, maximum_value,
                                                       increment FROM INFORMATION_SCHEMA.SEQUENCES WHERE sequence_schema || '.' || sequence_name
                                                       IN (SELECT pg_get_serial_sequence(':TABLE_NAME',':COLUMN_NAME'));");
        $seqNameStatement->execute(array(':TABLE_NAME' => $tableName, ':COLUMN_NAME' => $columnName));
        $seqName = $seqNameStatement->fetch(\PDO::FETCH_ASSOC);
        if(!$seqName) {
            throw Exception::sequenceDoesNotExist($tableName, $columnName);
        }
        return new DTO\Sequence($seqName["sequence_name"],$seqName["minimum_value"],$seqName["increment"],$seqName["maximum_value"]);
    }

}